from collections import namedtuple

import os
import re

from .listing import list_site_packages_paths


CAMEL_CASE_RE = re.compile(r'[A-Z][a-z0-9]+')
Argument = namedtuple('Argument', ['files', 'local_packages', 'directories'])


def package_type_order(future, stdlib, site, relative):
    order = 100
    if future:
        order = 0
    elif stdlib:
        order = 1
    elif site:
        order = 2
    elif not relative:
        order = 3
    return order


def condition_order(condition, relative):
    if not relative:
        order = 2
    elif not condition:
        order = 1
    else:
        order = 0
    return order


def canonical_sort_key(original_name, lineno, col_offset, relative,
                       local_package_names, distinguish_from_import=False):
    # Replace '_' (95) with '~' (128) to make it below 'z' (122)
    name = original_name.replace('_', '~')
    first_name = name.split('.', 1)[0]
    if relative:
        from_, _, variable = name.rpartition('.')
        from_ = from_.lower()
    else:
        from_ = ''
        variable = ''
    future = first_name == '~~future~~'
    local = (not first_name or
             first_name.replace('~', '_') in local_package_names)
    if future or local:
        site = False
        stdlib = False
    elif first_name in set(['flaskext']):
        # flaskext is created at runtime, and has no __file__
        site = True
        stdlib = False
    else:
        try:
            imported = __import__(first_name.replace('~', '_'))
        except ImportError as e:
            if not local_package_names:
                site = False
                stdlib = False
                local = True
            else:
                raise e
        else:
            site = getattr(
                imported,
                '__file__',
                ''
            ).startswith(tuple(list_site_packages_paths()))
            stdlib = not site
    sort_from_import = 1 if relative and distinguish_from_import else 0
    return (
        # FIXME: refactor it to use namedtuple
        # 1. Order: __future__, standard libraries, site-packages, local
        package_type_order(future, stdlib, site, relative),
        # 2. (optional) ``import ...`` is the first, ``from ... import ...` is
        # the next if ``distinguish_from_import`` is ``True``
        sort_from_import,
        from_ or first_name.lower(),
        # 3. CONSTANT_NAMES must be the first
        condition_order(variable.isupper(), relative),
        # 4. ClassNames must be the second
        condition_order(CAMEL_CASE_RE.search(variable), relative),
        # 5. Rest must be in alphabetical order
        name.lower()
    )


def sort_import_names(import_names, local_package_names,
                      distinguish_from_import=False):
    return sorted(
        import_names,
        key=lambda tup: canonical_sort_key(
            *tup,
            local_package_names=local_package_names,
            distinguish_from_import=distinguish_from_import
        )
    )


def sort_by_type(args):
    files = []
    local_packages = []
    directories = []
    for arg in args:
        name = arg.rstrip('/')
        try:
            __import__(name)
        except (ImportError, ValueError):
            if os.path.isdir(name):
                directories.append(name)
            elif os.path.exists(name):
                files.append(name)
            else:
                raise IOError("{0} dosen't exists.".format(name))
        else:
            local_packages.append(name)
    return Argument(files=files, local_packages=local_packages,
                    directories=directories)



from __future__ import print_function

import ast
import re
import sys

from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers.agile import PythonLexer

from .filter import Exclude
from .listing import list_all_argument, list_import_names
from .sort import canonical_sort_key, sort_by_type, sort_import_names


IGNORE_RE = re.compile(r'#\s*((fuck|shit)\s+)*ignore\s+import\s+order|'
                       r'#\s*no\s+import\s+order',
                       re.IGNORECASE)


def debug_import_names(import_names, local_package_names, highlight=None):
    fmt = '\x1b[30m{0}\x1b[39m \x1b[37m({1})\x1b[39m'.format
    hfmt = '\x1b[35m{0}\x1b[39m \x1b[37m({1})\x1b[39m'.format
    return ', '.join(
        (hfmt if highlight in (tup, tup[0]) else fmt)(
            tup[0],
            ','.join(str(v)
                     for v in canonical_sort_key(
                         *tup, local_package_names=local_package_names))
        )
        for tup in import_names
    )


def inspect_order(args, debug, only_file=False, excludes=[],
                  distinguish_from_import=False):
    argument = sort_by_type(args)
    if not argument.local_packages and not only_file:
        raise ValueError('At least 1 local package name required.')
    filters = []
    if excludes:
        filters.append(Exclude(excludes))
    files = list_all_argument(argument, filters=filters)
    errored = False
    for filename in files:
        with open(filename) as file_:
            if IGNORE_RE.search('\n'.join(file_.readline() for _ in range(3))):
                continue
            file_.seek(0)
            tree = ast.parse(file_.read(), filename)
        import_names = list(list_import_names(tree))
        canonical_order = sort_import_names(import_names,
                                            argument.local_packages,
                                            distinguish_from_import)
        prev_import = None
        for actual, expected in zip(import_names, canonical_order):
            if actual[0] != expected[0]:
                errored = True
                code_offset = min(expected[1], actual[1])
                code_end = max(expected[1], actual[1])
                print(
                    '\x1b[35m{0}\x1b[39m:{1}-{2}:'.format(
                        filename, code_offset, code_end
                    ),
                    end=' ',
                    file=sys.stderr
                )
                if prev_import is None:  # first
                    print(expected[0], 'must be the first, not', actual[0],
                          file=sys.stderr)
                else:
                    print(expected[0], 'must be above than', actual[0],
                          file=sys.stderr)
                lineno_cols = len(str(code_end))
                format_line = (u'{0:' + str(lineno_cols) + '} {1}').format
                with open(filename, 'rb') as file_:
                    highlighted = highlight(
                        file_.read(),
                        PythonLexer(),
                        Terminal256Formatter()
                    )
                sliced = highlighted.splitlines()[code_offset - 1:code_end]
                codelisting = '\n'.join(
                    format_line(i + code_offset, line)
                    for i, line in enumerate(sliced)
                )
                print(codelisting, file=sys.stderr)
                if debug:
                    print('\x1b[32;49;1mExpected order:\x1b[39;49;00m',
                          debug_import_names(canonical_order,
                                             argument.local_packages,
                                             expected),
                          file=sys.stderr)
                    print('\x1b[31;49;1mActual order:\x1b[39;49;00m  ',
                          debug_import_names(import_names,
                                             argument.local_packages,
                                             actual),
                          file=sys.stderr)
                print(file=sys.stderr)
                break
            prev_import = actual

    if errored:
        raise SystemExit(1)



from pytest import raises
from import_order.inspect import debug_import_names, inspect_order


def test_inspect_order_raise_without_local_packages():
    with raises(ValueError):
        inspect_order(['.'], False)


def test_debug_import_names():
    debug_names = debug_import_names(
        [('re', 0, 0, False), ('datetime', 0, 0, False)],
        ['import_order'],
        ('re', 0, 0, False)
    )
    fmt = '\x1b[30m{0}\x1b[39m \x1b[37m({1})\x1b[39m'.format(
        'datetime', '1,0,datetime,2,2,datetime'
    )
    hfmt = '\x1b[35m{0}\x1b[39m \x1b[37m({1})\x1b[39m'.format(
        're', '1,0,re,2,2,re'
    )
    assert debug_names == ', '.join([hfmt, fmt])



from pytest import raises

from import_order.sort import (canonical_sort_key,
                               sort_by_type, sort_import_names)


def test_sort_by_type():
    files = ['tests/mock.py']
    directories = ['.', 'tests/test_dir']
    packages = ['tests']
    arguments = files + directories + packages
    argument = sort_by_type(arguments)
    assert set(argument.files) == set(files)
    assert set(argument.local_packages) == set(packages)
    assert set(argument.directories) == set(directories)
    with raises(IOError):
        sort_by_type(['foobar/abc.py'])


def test_sort_relative_import_belower():
    local_package_names = ['hello']
    import_names = [('hello.b', 1, 0, False), ('.a', 2, 0, True)]
    sorted_ = sort_import_names(import_names, local_package_names)
    assert import_names == sorted_


def test_sort_constant():
    local_package_names = ['hello']
    import_names = [('.FOO', 1, 0, True), ('.foo', 2, 0, True)]
    sorted_ = sort_import_names(import_names, local_package_names)
    assert import_names == sorted_


def test_sort_class_name():
    local_package_names = ['hello']
    import_names = [('.FooClass', 1, 0, True), ('.foo', 2, 0, True)]
    sorted_ = sort_import_names(import_names, local_package_names)
    assert import_names == sorted_


def test_sort_all():
    local_package_names = ['hello']
    import_names = [
        ('re.compile', 1, 0, False),
        ('pygments.highlight', 2, 0, False),
        ('hello.FOO', 3, 0, False),
        ('hello.foo', 3, 0, False),
        ('.FooClass', 4, 0, True),
        ('.a', 4, 0, True),
    ]
    sorted_ = sort_import_names(import_names, local_package_names)
    assert import_names == sorted_


def test_sort_distinguish():
    local_package_names = ['hello']
    import_names = [
        ('re.compile', 1, 0, False),
        ('pytest', 2, 0, False),
        ('pygments.highlight', 2, 0, True),
        ('hello.FOO', 3, 0, False),
        ('hello.foo', 3, 0, False),
        ('.FooClass', 4, 0, True),
        ('.a', 4, 0, True),
    ]
    sorted_ = sort_import_names(import_names, local_package_names, True)
    assert import_names == sorted_


def test_canonical_sort_key():
    # ensure re is stdlib
    actual = canonical_sort_key(
        're.compile', 1, 0, False, local_package_names=['hello']
    )
    assert (1, 0, 're', 2, 2, 're.compile') == actual
    # ensure pygments is site packages
    actual = canonical_sort_key(
        'pygments', 1, 0, False, local_package_names=['hello']
    )
    assert (2, 0, 'pygments', 2, 2, 'pygments') == actual
    # ensure hello is local packages
    actual = canonical_sort_key(
        'hello', 1, 0, False, local_package_names=['hello']
    )
    assert (3, 0, 'hello', 2, 2, 'hello') == actual



import os


class FilterMixin(object):

    def apply(self):
        raise NotImplementedError()


class Exclude(FilterMixin):

    def __init__(self, excludes):
        excludes = [os.path.abspath(path) for path in excludes]
        self.file_excluds = [path for path in excludes if os.path.isfile(path)]
        self.dir_excluds = ['{path}{sep}'.format(path=path, sep=os.sep)
                            for path in excludes if os.path.isdir(path)]

    def apply(self, files):
        results = []
        for filename in files:
            abs_name = os.path.abspath(filename)
            exclude = False
            if abs_name in self.file_excluds:
                exclude = True
            for path in self.dir_excluds:
                if abs_name.startswith(path):
                    exclude = True
                    break
            if not exclude:
                results.append(filename)
        return results



import ast
import os
import os.path
import site
import sys


def list_site_packages_paths():
    site_packages_paths = set([site.USER_SITE])
    try:
        site_packages_paths.update(site.getsitepackages())
    except AttributeError:
        pass
    try:
        user_site = site.getusersitepackages()
        if isinstance(user_site, str):
            site_packages_paths.add(user_site)
        else:
            site_packages_paths.update(user_site)
    except AttributeError:
        pass
    try:
        virtualenv_path = os.environ['VIRTUAL_ENV']
    except KeyError:
        pass
    else:
        virtualenv_src_path = os.path.join(virtualenv_path, 'src')
        site_packages_paths.update(
            path
            for path in sys.path
            if path.startswith(virtualenv_path) and (
                'site-packages' in path or
                path.startswith(virtualenv_src_path)
            )
        )
    return site_packages_paths


def list_python_files(dirname):
    for entry in os.listdir(dirname):
        path = os.path.join(dirname, entry)
        if os.path.isdir(path):
            for subentry in list_python_files(path):
                yield subentry
        elif entry.endswith('.py'):
            yield path


def list_import_names(tree):
    for node in ast.iter_child_nodes(tree):
        if isinstance(node, ast.Import):
            for name in node.names:
                yield name.name, node.lineno, node.col_offset, False
        elif isinstance(node, ast.ImportFrom):
            for name in node.names:
                yield ('.' * node.level +
                       (node.module + '.' if node.module else '') +
                       name.name,
                       node.lineno, node.col_offset, True)
        else:
            continue


def list_all_argument(argument, filters=[]):
    files = argument.files
    for local_package_name in argument.local_packages:
        files.extend(list_python_files(local_package_name))
    for directory in argument.directories:
        files.extend(list_python_files(directory))
    for filter_ in filters:
        files = filter_.apply(files)
    return set(files)



from setuptools import find_packages, setup

from import_order.version import VERSION


install_requires = [
    'Pygments >= 2.0.2, < 2.1.0',
]


tests_require = [
    'pytest >= 2.7.0',
    'tox',
]


def readme():
    with open('README.rst') as f:
        try:
            return f.read()
        except (IOError, OSError):
            return None


setup(
    name="import-order",
    version=VERSION,
    packages=find_packages(),
    url='https://github.com/spoqa/import-order',
    install_requires=install_requires,
    tests_require=tests_require,
    extras_require={
        'tests': tests_require,
    },
    maintainer='Spoqa',
    maintainer_email='dev' '@' 'spoqa.com',
    entry_points={
        'console_scripts': [
            'import-order = import_order.cli:main',
        ]
    },
    description='Check python import order.',
    long_description=readme(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Operating System :: Unix',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
    ]
)



import optparse

from .inspect import inspect_order
from .version import VERSION


parser = optparse.OptionParser(
    usage='%prog [options] [local package ...] [file ...] [directory ...]',
    version='%prog {0}'.format(VERSION))
parser.add_option('-d', '--debug', action='store_true', default=False,
                  help='Debug mode.')
parser.add_option('--only-file', action='store_true', default=False,
                  help='Inspect only files not packages')
parser.add_option('--exclude', action='append', default=[],
                  help='Ignore specific file or directory.')
parser.add_option('--distinguish-from-import', action='append', default=[],
                  help='Distinguish `from ... import ...` and `import ...`')


def main():
    options, args = parser.parse_args()
    inspect_order(args, options.debug, options.only_file, options.exclude)



import os

from import_order.filter import Exclude
from import_order.listing import list_all_argument
from import_order.sort import sort_by_type


def test_filter_exclude():
    files = list_all_argument(sort_by_type(['tests']))

    exclude_filter = Exclude([])
    assert 'tests/mock.py' in exclude_filter.apply(files)
    exclude_filter = Exclude(['mock.py'])
    assert 'tests/mock.py' in exclude_filter.apply(files)

    exclude_filter = Exclude(['.'])
    assert exclude_filter.apply(files) == []
    exclude_filter = Exclude(['tests'])
    assert exclude_filter.apply(files) == []
    exclude_filter = Exclude(['tests/../tests'])
    assert exclude_filter.apply(files) == []
    exclude_filter = Exclude([os.path.abspath('tests')])
    assert exclude_filter.apply(files) == []

    exclude_filter = Exclude(['tests/test_dir/foo'])
    assert set(files) == set(exclude_filter.apply(files))

    files = ['tests/test_dir/foo.py', 'tests/test_dir/bar.py',
             'tests/mock.py']
    exclude_filter = Exclude(['tests/mock.py'])
    assert set(exclude_filter.apply(files)) == set(['tests/test_dir/foo.py',
                                                    'tests/test_dir/bar.py'])
    exclude_filter = Exclude(['tests/test_dir'])
    assert set(exclude_filter.apply(files)) == set(['tests/mock.py'])
    exclude_filter = Exclude(['tests/test_dir/foo.py',
                              'tests/test_dir/bar.py'])
    assert set(exclude_filter.apply(files)) == set(['tests/mock.py'])
